==> peekabot-0.6.0.tar.bz2: peekabot-0.6.0/src/atl/ClientConnection.cc <==
/*
 * Copyright Staffan Gimåker 2006-2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cstdio>
#include <numeric>
#include <boost/format.hpp>

#include "ClientConnection.hh"
#include "ConnectionStates.hh"
#include "ActionTransporter.hh"
#include "../serialization/SerializationInterface.hh"
#include "../serialization/ChunkedBufferAdapter.hh"
#include "../MessageHub.hh"

//
// Adjust these (upwards) to squeeze a higher maximum throughput out
// of the system - you pay the cost in memory.
//
static const size_t INTERMEDIARY_BUFFER_SIZE = 32*1024; // in bytes
static const size_t SEND_BUFFER_CHUNK_SIZE   = 4*1024;  // in bytes


using namespace peekabot;
using namespace peekabot::atl;


ClientConnection::ClientConnection(
    ActionTransporter *transporter, 
    sockets::SocketType socket_fd,
    SourceID client_id) throw()
    : m_intermediary_buf(INTERMEDIARY_BUFFER_SIZE),
      m_send_buf(SEND_BUFFER_CHUNK_SIZE),
      m_socket_fd(socket_fd),
      m_transporter(transporter),
      m_state(0),
      m_id(client_id),
      m_is_closed(false),
      m_disconnect_reason(LOCALLY_DISCONNECTED),
      m_sent_throughput(0),
      m_recv_throughput(0),
      m_bytes_sent(0),
      m_bytes_recv(0),
      m_sent_since_last_tick(0),
      m_recv_since_last_tick(0),
      m_up_since(boost::posix_time::microsec_clock::local_time()),
      m_low_latency_mode(false)
{
    sockaddr_in name;
    socklen_t namelen = sizeof(name);
    if( getpeername(m_socket_fd, (sockaddr *)&name, &namelen) == SOCKET_ERROR )
    {
        // TODO: report error   
        m_remote_address = "???";
    }
    else
    {
        m_remote_address = inet_ntoa(name.sin_addr);
    }



    set_state(new UnauthenticatedState(this));

    // Make the socket unblocking
    sockets::set_nonblocking(m_socket_fd);

    // Notify observers we're up and running
    notify(ON_CLIENT_CONNECTED);
}


ClientConnection::~ClientConnection() throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    if( is_closed() )
        close();
}


bool ClientConnection::is_closed() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    return m_is_closed;
}


void ClientConnection::close(
    ClientConnection::DisconnectReason reason) throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);
    
    m_disconnect_reason = reason;
    
    set_state(0);
    sockets::close_socket(m_socket_fd);
    connection_closed(false);
}


void ClientConnection::connection_closed(bool remotely_closed) throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    if( !m_is_closed )
    {
        m_is_closed = true;
        notify(ON_CLIENT_DISCONNECTED);

        if( remotely_closed )
            m_disconnect_reason = REMOTELY_DISCONNECTED;
        // else: close takes care of setting the disconnect reason
    }
}


ClientConnection::DisconnectReason 
ClientConnection::get_disconnect_reason() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    return m_disconnect_reason;
}


size_t ClientConnection::get_outbound_count() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    return m_intermediary_buf.get_bytes_available() + m_send_buf.get_size();
}


boost::posix_time::time_duration ClientConnection::get_uptime() const throw()
{
    return boost::posix_time::microsec_clock::local_time() - m_up_since;
}


std::pair<float, float> ClientConnection::get_throughput() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_stat_mutex);

    return std::make_pair(m_sent_throughput, m_recv_throughput);
}


uint64_t ClientConnection::get_bytes_sent() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_stat_mutex);

    return m_bytes_sent;
}


uint64_t ClientConnection::get_bytes_recv() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_stat_mutex);

    return m_bytes_recv;
}



std::string ClientConnection::get_remote_address() const throw()
{
    return m_remote_address;
}



void ClientConnection::send(const void *buf, size_t bytes) throw()
{
    if( get_outbound_count() < 1 && bytes > 0 )
        set_active_sending();

    {
        boost::recursive_mutex::scoped_lock lock(m_mutex);

        m_send_buf.write(buf, bytes);
    }
}



void ClientConnection::send(ChunkedBuffer &buf) throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    static uint8_t ibuf[SEND_BUFFER_CHUNK_SIZE];

    while( !buf.is_empty() )
    {
        size_t n = buf.read(ibuf, SEND_BUFFER_CHUNK_SIZE);
        send(ibuf, n);
    }
}


void ClientConnection::send_action(Action *action) throw()
{
    {
        boost::recursive_mutex::scoped_lock lock(m_mutex);

        if( get_outbound_count() < 1 )
            set_active_sending();

        uint32_t body_size  = 0xDEADBEEF;

        // Store the position in the send buffer at which the body
        // size dword is stored, see below!
        size_t size_pos = m_send_buf.get_size();

        // Since we don't know the size of the action body a priori we
        // write a placeholder to the send buffer and overwrite it once
        // we've obtained the relevant information.
        m_send_buf.write(&body_size, sizeof(uint32_t));

        // Store the size of the send buffer prior to serializing 
        // the action body
        size_t s1 = m_send_buf.get_size();

        ChunkedBufferAdapter adapter(m_send_buf);
        SerializationInterface ser_buf(adapter);

        try
        {
            ser_buf << action;
        }
        catch(std::exception &e)
        {
            TheMessageHub::instance().publish(
                ERROR_MESSAGE,
                "Failed to serialize action, exception caught: %s", e.what());
            return;
        }
        catch(...)
        {
            TheMessageHub::instance().publish(
                ERROR_MESSAGE,
                "Failed to serialize action, unknown exception caught");
            return;
        }

        // Calculate the body size -- that is, the number of bytes the send 
        // buffer grew during serialization.
        body_size  = m_send_buf.get_size() - s1;
        
        // Overwrite the old placeholder size value
        m_send_buf.overwrite(
            &body_size, sizeof(uint32_t), size_pos);
    }
}






size_t ClientConnection::recv(void *buf, size_t max_bytes) throw()
{
    if( max_bytes < 1 )
        return 0;

    ssize_t n = 0;

    n = ::recv(m_socket_fd, (sockets::BufType)buf, max_bytes, 0);

    {
        boost::recursive_mutex::scoped_lock lock(m_stat_mutex);
        
        m_bytes_recv += n;
        m_recv_since_last_tick += n;
    }
    
    if( n < 1 )
    {
        connection_closed(true);

        return 0;
    }

    return n;
}


size_t ClientConnection::peek(void *buf, size_t max_bytes) throw()
{
    if( max_bytes < 1 )
        return 0;

    ssize_t n = 0;

    n = ::recv(m_socket_fd, (sockets::BufType)buf, max_bytes, MSG_PEEK);

    return n;
}

    
void ClientConnection::notify(ConnectionEvent event) throw()
{
    m_transporter->notify(event, this);
}


void ClientConnection::set_state(ConnectionState *state) throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    if( m_state != state )
    {
        if( m_state != 0 )
        {
            m_state->on_state_inactive();

            // Delete old state
            delete m_state;
        }

        if( state != 0 )
            state->on_activating_state();
        
        m_state = state;
        if( m_state != 0 )
            m_state->on_state_active();
    }
}


SourceID ClientConnection::get_id() const throw()
{
    return m_id;
}


ActionTransporter *ClientConnection::get_transporter() throw()
{
    return m_transporter;
}


const ActionTransporter *ClientConnection::get_transporter() const throw()
{
    return m_transporter;
}


void ClientConnection::on_tick(boost::posix_time::time_duration &dt) throw()
{
    boost::recursive_mutex::scoped_lock lock(m_stat_mutex);

    // 1. update throughput!
    m_sent_throughput = m_sent_since_last_tick / 
        (dt.total_milliseconds() / 1000.0f);
    m_recv_throughput = m_recv_since_last_tick / 
        (dt.total_milliseconds() / 1000.0f);

    TheMessageHub::instance().publish(
        DEBUG_MESSAGE,
        "Client %d throughput (%s): %f / %f Kb/s", 
        m_id,
        get_remote_address().c_str(),
        m_sent_throughput/1024,
        m_recv_throughput/1024);
        

    /*std::cout << get_remote_address() << ": "
              << m_sent_throughput/1024/1024 << " / " 
              << m_recv_throughput/1024/1024 << " Mb/s" << std::endl;*/

    m_sent_since_last_tick = m_recv_since_last_tick = 0;

    {
        boost::recursive_mutex::scoped_lock lock(m_mutex);

        // 2. call on_tick on the state!
        if( m_state )
            m_state->on_tick(dt);
    }

}


void ClientConnection::on_data_available() throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    if( m_state )
        m_state->on_data_available();
}

  
sockets::SocketType ClientConnection::get_socket_fd() const throw()
{
    return m_socket_fd;
}


void ClientConnection::set_active_sending() throw()
{
    notify(ON_BUSY_SENDING);

    //
    // Wake up the transporter so we can start sending right away.
    //
    get_transporter()->wake();
}


void ClientConnection::set_idle_sending() throw()
{
    notify(ON_IDLE_SENDING);
}


void ClientConnection::on_writable() throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    //if( is_closed() )
    //    return;

    // no data left to write?
    if( get_outbound_count() == 0 && m_state )
        // request new data from state
        m_state->on_send_buffers_exhausted();

    // no data in the intermediary buffer?
    if( m_intermediary_buf.get_bytes_available() == 0 )
    {
        m_intermediary_buf.clear();

        // see if there's data available to fill it up
        if( !m_send_buf.is_empty() )
        {
            m_intermediary_buf.m_wptr += m_send_buf.read(
                m_intermediary_buf.m_buf, 
                m_intermediary_buf.m_capacity);
        }
    }

    // are there any bytes available for sending?
    if( m_intermediary_buf.get_bytes_available() > 0 )
    {
        // then send 'em!
        size_t sent = ::send(
            m_socket_fd, 
            (sockets::BufType)m_intermediary_buf.get_read_ptr(),
            m_intermediary_buf.get_bytes_available(), 0);

        if( sent < 1 )
            connection_closed(true);

        m_intermediary_buf.m_rptr += sent;

        m_bytes_sent += sent;
        m_sent_since_last_tick += sent;
    }
    else
        // we'll not be sending anything just now
        set_idle_sending();
}



uint32_t ClientConnection::get_protocol_version() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    return m_protocol_version;
}


std::string ClientConnection::get_proxy_version() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    return (boost::format("%1%.%2%.%3%") % 
            m_proxy_version_major % 
            m_proxy_version_minor % 
            m_proxy_version_revision).str();
}


uint8_t ClientConnection::get_proxy_version_major() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    return m_proxy_version_major;
}


uint8_t ClientConnection::get_proxy_version_minor() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    return m_proxy_version_minor;
}


uint8_t ClientConnection::get_proxy_version_revision() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    return m_proxy_version_revision;
}


void ClientConnection::set_protocol_version(uint32_t version) throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    m_protocol_version = version;
}


void ClientConnection::set_proxy_version(
    uint8_t major, uint8_t minor, uint8_t revision) throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    m_proxy_version_major = major;
    m_proxy_version_minor = minor;
    m_proxy_version_revision = revision;
}



/*ConnectionState *ClientConnection::use_state() throw()
{
    return m_state;
}*/


void ClientConnection::toggle_low_latency_mode(bool enable) throw()
{
    if( m_low_latency_mode != enable )
    {
        m_low_latency_mode = enable;

        int toggle = enable ? 1:0;
        if( setsockopt(get_socket_fd(), IPPROTO_TCP, TCP_NODELAY, 
                       (sockets::BufType)&toggle, sizeof(int)) == SOCKET_ERROR )
            perror("Warning: setsockopt failed to toggle TCP_NODELAY");
    }
}


bool ClientConnection::low_latency_mode_enabled() const throw()
{
    return m_low_latency_mode;
}




//
// --------- ConnectionState impl. -----------------
//


ClientConnection *ConnectionState::get_connection() throw()
{
    return m_conn;
}


const ClientConnection *ConnectionState::get_connection() const throw()
{
    return m_conn;
}


void ConnectionState::notify_connection(ConnectionEvent event) throw()
{
    m_conn->notify(event);
}


void ConnectionState::set_proxy_version(
    uint8_t major, uint8_t minor, uint8_t rev) throw()
{
    m_conn->set_proxy_version(major, minor, rev);
}


void ConnectionState::set_protocol_version(uint32_t version) throw()
{
    m_conn->set_protocol_version(version);
}

==> peekabot-0.7.2.tar.bz2: peekabot-0.7.2/src/atl/ClientConnection.cc <==
/*
 * Copyright Staffan Gimåker 2006-2009.
 *
 * ---
 *
 * This file is part of peekabot.
 *
 * peekabot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * peekabot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cstdio>
#include <numeric>
#include <boost/format.hpp>

#include "ClientConnection.hh"
#include "ConnectionStates.hh"
#include "ActionTransporter.hh"
#include "../serialization/SerializationInterface.hh"
#include "../serialization/ChunkedBufferAdapter.hh"
#include "../MessageHub.hh"
#include "../Version.hh"

//
// Adjust these (upwards) to squeeze a higher maximum throughput out
// of the system - you pay the cost in memory.
//
static const size_t INTERMEDIARY_BUFFER_SIZE = 32*1024; // in bytes
static const size_t SEND_BUFFER_CHUNK_SIZE   = 4*1024;  // in bytes


using namespace peekabot;
using namespace peekabot::atl;


ClientConnection::ClientConnection(
    ActionTransporter *transporter, 
    sockets::SocketType socket_fd,
    SourceID client_id) throw()
    : m_intermediary_buf(INTERMEDIARY_BUFFER_SIZE),
      m_send_buf(SEND_BUFFER_CHUNK_SIZE),
      m_socket_fd(socket_fd),
      m_transporter(transporter),
      m_state(0),
      m_id(client_id),
      m_is_closed(false),
      m_disconnect_reason(LOCALLY_DISCONNECTED),
      m_sent_throughput(0),
      m_recv_throughput(0),
      m_bytes_sent(0),
      m_bytes_recv(0),
      m_sent_since_last_tick(0),
      m_recv_since_last_tick(0),
      m_up_since(boost::posix_time::microsec_clock::local_time()),
      m_low_latency_mode(false)
{
    sockaddr_in name;
    socklen_t namelen = sizeof(name);
    if( getpeername(m_socket_fd, (sockaddr *)&name, &namelen) == SOCKET_ERROR )
    {
        // TODO: report error   
        m_remote_address = "???";
    }
    else
    {
        m_remote_address = inet_ntoa(name.sin_addr);
    }



    set_state(new UnauthenticatedState(this));

    // Make the socket unblocking
    sockets::set_nonblocking(m_socket_fd);

    // Notify observers we're up and running
    notify(ON_CLIENT_CONNECTED);
}


ClientConnection::~ClientConnection() throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    if( is_closed() )
        close();
}


bool ClientConnection::is_closed() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    return m_is_closed;
}


void ClientConnection::close(
    ClientConnection::DisconnectReason reason) throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);
    
    m_disconnect_reason = reason;
    
    set_state(0);
    sockets::close_socket(m_socket_fd);
    connection_closed(false);
}


void ClientConnection::connection_closed(bool remotely_closed) throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    if( !m_is_closed )
    {
        m_is_closed = true;
        notify(ON_CLIENT_DISCONNECTED);

        if( remotely_closed )
            m_disconnect_reason = REMOTELY_DISCONNECTED;
        // else: close takes care of setting the disconnect reason
    }
}


ClientConnection::DisconnectReason 
ClientConnection::get_disconnect_reason() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    return m_disconnect_reason;
}


size_t ClientConnection::get_outbound_count() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    return m_intermediary_buf.get_bytes_available() + m_send_buf.get_size();
}


boost::posix_time::time_duration ClientConnection::get_uptime() const throw()
{
    return boost::posix_time::microsec_clock::local_time() - m_up_since;
}


std::pair<float, float> ClientConnection::get_throughput() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_stat_mutex);

    return std::make_pair(m_sent_throughput, m_recv_throughput);
}


uint64_t ClientConnection::get_bytes_sent() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_stat_mutex);

    return m_bytes_sent;
}


uint64_t ClientConnection::get_bytes_recv() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_stat_mutex);

    return m_bytes_recv;
}



std::string ClientConnection::get_remote_address() const throw()
{
    return m_remote_address;
}



void ClientConnection::send(const void *buf, size_t bytes) throw()
{
    if( get_outbound_count() < 1 && bytes > 0 )
        set_active_sending();

    {
        boost::recursive_mutex::scoped_lock lock(m_mutex);

        m_send_buf.write(buf, bytes);
    }
}



void ClientConnection::send(ChunkedBuffer &buf) throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    static uint8_t ibuf[SEND_BUFFER_CHUNK_SIZE];

    while( !buf.is_empty() )
    {
        size_t n = buf.read(ibuf, SEND_BUFFER_CHUNK_SIZE);
        send(ibuf, n);
    }
}


void ClientConnection::send_action(Action *action) throw()
{
    {
        boost::recursive_mutex::scoped_lock lock(m_mutex);

        if( get_outbound_count() < 1 )
            set_active_sending();

        uint32_t body_size  = 0xDEADBEEF;

        // Store the position in the send buffer at which the body
        // size dword is stored, see below!
        size_t size_pos = m_send_buf.get_size();

        // Since we don't know the size of the action body a priori we
        // write a placeholder to the send buffer and overwrite it once
        // we've obtained the relevant information.
        m_send_buf.write(&body_size, sizeof(uint32_t));

        // Store the size of the send buffer prior to serializing 
        // the action body
        size_t s1 = m_send_buf.get_size();

        ChunkedBufferAdapter adapter(m_send_buf);
        SerializationInterface ser_buf(adapter);

        try
        {
            ser_buf << action;
        }
        catch(std::exception &e)
        {
            TheMessageHub::instance().publish(
                ERROR_MESSAGE,
                "Failed to serialize action, exception caught: %s", e.what());
            return;
        }
        catch(...)
        {
            TheMessageHub::instance().publish(
                ERROR_MESSAGE,
                "Failed to serialize action, unknown exception caught");
            return;
        }

        // Calculate the body size -- that is, the number of bytes the send 
        // buffer grew during serialization.
        body_size  = m_send_buf.get_size() - s1;
        
        // Overwrite the old placeholder size value
        m_send_buf.overwrite(
            &body_size, sizeof(uint32_t), size_pos);
    }
}






size_t ClientConnection::recv(void *buf, size_t max_bytes) throw()
{
    if( max_bytes < 1 )
        return 0;

    ssize_t n = 0;

    n = ::recv(m_socket_fd, (sockets::BufType)buf, max_bytes, 0);

    {
        boost::recursive_mutex::scoped_lock lock(m_stat_mutex);
        
        m_bytes_recv += n;
        m_recv_since_last_tick += n;
    }
    
    if( n < 1 )
    {
        connection_closed(true);

        return 0;
    }

    return n;
}


size_t ClientConnection::peek(void *buf, size_t max_bytes) throw()
{
    if( max_bytes < 1 )
        return 0;

    ssize_t n = 0;

    n = ::recv(m_socket_fd, (sockets::BufType)buf, max_bytes, MSG_PEEK);

    return n;
}

    
void ClientConnection::notify(ConnectionEvent event) throw()
{
    m_transporter->notify(event, this);
}


void ClientConnection::set_state(ConnectionState *state) throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    if( m_state != state )
    {
        if( m_state != 0 )
        {
            m_state->on_state_inactive();

            // Delete old state
            delete m_state;
        }

        if( state != 0 )
            state->on_activating_state();
        
        m_state = state;
        if( m_state != 0 )
            m_state->on_state_active();
    }
}


SourceID ClientConnection::get_id() const throw()
{
    return m_id;
}


ActionTransporter *ClientConnection::get_transporter() throw()
{
    return m_transporter;
}


const ActionTransporter *ClientConnection::get_transporter() const throw()
{
    return m_transporter;
}


void ClientConnection::on_tick(boost::posix_time::time_duration &dt) throw()
{
    boost::recursive_mutex::scoped_lock lock(m_stat_mutex);

    // 1. update throughput!
    m_sent_throughput = m_sent_since_last_tick / 
        (dt.total_milliseconds() / 1000.0f);
    m_recv_throughput = m_recv_since_last_tick / 
        (dt.total_milliseconds() / 1000.0f);

    TheMessageHub::instance().publish(
        DEBUG_MESSAGE,
        "Client %d throughput (%s): %f / %f Kb/s", 
        m_id,
        get_remote_address().c_str(),
        m_sent_throughput/1024,
        m_recv_throughput/1024);
        

    /*std::cout << get_remote_address() << ": "
              << m_sent_throughput/1024/1024 << " / " 
              << m_recv_throughput/1024/1024 << " Mb/s" << std::endl;*/

    m_sent_since_last_tick = m_recv_since_last_tick = 0;

    {
        boost::recursive_mutex::scoped_lock lock(m_mutex);

        // 2. call on_tick on the state!
        if( m_state )
            m_state->on_tick(dt);
    }

}


void ClientConnection::on_data_available() throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    if( m_state )
        m_state->on_data_available();
}

  
sockets::SocketType ClientConnection::get_socket_fd() const throw()
{
    return m_socket_fd;
}


void ClientConnection::set_active_sending() throw()
{
    notify(ON_BUSY_SENDING);

    //
    // Wake up the transporter so we can start sending right away.
    //
    get_transporter()->wake();
}


void ClientConnection::set_idle_sending() throw()
{
    notify(ON_IDLE_SENDING);
}


void ClientConnection::on_writable() throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    //if( is_closed() )
    //    return;

    // no data left to write?
    if( get_outbound_count() == 0 && m_state )
        // request new data from state
        m_state->on_send_buffers_exhausted();

    // no data in the intermediary buffer?
    if( m_intermediary_buf.get_bytes_available() == 0 )
    {
        m_intermediary_buf.clear();

        // see if there's data available to fill it up
        if( !m_send_buf.is_empty() )
        {
            m_intermediary_buf.m_wptr += m_send_buf.read(
                m_intermediary_buf.m_buf, 
                m_intermediary_buf.m_capacity);
        }
    }

    // are there any bytes available for sending?
    if( m_intermediary_buf.get_bytes_available() > 0 )
    {
        // then send 'em!
        size_t sent = ::send(
            m_socket_fd, 
            (sockets::BufType)m_intermediary_buf.get_read_ptr(),
            m_intermediary_buf.get_bytes_available(), 0);

        if( sent < 1 )
            connection_closed(true);

        m_intermediary_buf.m_rptr += sent;

        m_bytes_sent += sent;
        m_sent_since_last_tick += sent;
    }
    else
        // we'll not be sending anything just now
        set_idle_sending();
}



uint32_t ClientConnection::get_protocol_version() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    return m_protocol_version;
}


std::string ClientConnection::get_client_version() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);
    return make_peekabot_version_string(
        m_client_version_major,
        m_client_version_minor,
        m_client_version_revision,
        m_client_version_rc);
}


uint8_t ClientConnection::get_client_version_major() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    return m_client_version_major;
}


uint8_t ClientConnection::get_client_version_minor() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    return m_client_version_minor;
}


uint8_t ClientConnection::get_client_version_revision() const throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    return m_client_version_revision;
}


void ClientConnection::set_protocol_version(uint32_t version) throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    m_protocol_version = version;
}


void ClientConnection::set_client_version(
    uint8_t major, uint8_t minor, uint8_t revision, uint8_t rc) throw()
{
    boost::recursive_mutex::scoped_lock lock(m_mutex);

    m_client_version_major = major;
    m_client_version_minor = minor;
    m_client_version_revision = revision;
    m_client_version_rc = rc;
}



/*ConnectionState *ClientConnection::use_state() throw()
{
    return m_state;
}*/


void ClientConnection::toggle_low_latency_mode(bool enable) throw()
{
    if( m_low_latency_mode != enable )
    {
        m_low_latency_mode = enable;

        int toggle = enable ? 1:0;
        if( setsockopt(get_socket_fd(), IPPROTO_TCP, TCP_NODELAY, 
                       (sockets::BufType)&toggle, sizeof(int)) == SOCKET_ERROR )
            perror("Warning: setsockopt failed to toggle TCP_NODELAY");
    }
}


bool ClientConnection::low_latency_mode_enabled() const throw()
{
    return m_low_latency_mode;
}




//
// --------- ConnectionState impl. -----------------
//


ClientConnection *ConnectionState::get_connection() throw()
{
    return m_conn;
}


const ClientConnection *ConnectionState::get_connection() const throw()
{
    return m_conn;
}


void ConnectionState::notify_connection(ConnectionEvent event) throw()
{
    m_conn->notify(event);
}


void ConnectionState::set_client_version(
    uint8_t major, uint8_t minor, uint8_t rev, uint8_t rc) throw()
{
    m_conn->set_client_version(major, minor, rev, rc);
}


void ConnectionState::set_protocol_version(uint32_t version) throw()
{
    m_conn->set_protocol_version(version);
}
